# This file is placed in the Public Domain.


"classes"


import inspect


from .disk    import Workdir
from .object  import Object


class Classes: # pylint: disable=R0903

    "Whitelist"

    classes = Object()

    @staticmethod
    def long(name):
        "match from single name to long name."
        split = name.split(".")[-1].lower()
        res = name
        for named in Classes.classes:
            if split in named.split(".")[-1].lower():
                res = named
                break
        if "." not in res:
            for fnm in Workdir.types():
                claz = fnm.split(".")[-1]
                if fnm == claz.lower():
                    res = fnm
        return res

    @staticmethod
    def whitelist(clz):
        "add class to whitelist."
        name = str(clz).split()[1][1:-2]
        setattr(Classes.classes, name, clz)


def scancls(mod) -> None:
    "scan module for classes."
    for key, clz in inspect.getmembers(mod, inspect.isclass):
        if key.startswith("cb"):
            continue
        if not issubclass(clz, Object):
            continue
        Classes.whitelist(clz)


def __dir__():
    return (
        "Classes",
        "scancls"
    )



# This file is placed in the Public Domain.


"outline processor markup language"


from ..default import Default
from ..object  import construct, update
from ..run     import broker
from ..utils   import shortid, spl


from .rss import Rss


TEMPLATE = """<opml version="1.0">
    <head>
        <title>rssbot opml</title>
    </head>
    <body>
        <outline title="rssbot opml" text="24/7 feed fetcher">"""


class Parser:

    "Parser"

    @staticmethod
    def getnames(line):
        "return list of attribute names."
        return [x.split('="')[0]  for x in line.split()]

    @staticmethod
    def getvalue(line, attr):
        "retrieve attribute value."
        lne = ''
        index1 = line.find(f'{attr}="')
        if index1 == -1:
            return lne
        index1 += len(attr) + 2
        index2 = line.find('"', index1)
        if index2 == -1:
            index2 = line.find('/>', index1)
        if index2 == -1:
            return lne
        lne = line[index1:index2]
        if 'CDATA' in lne:
            lne = lne.replace('![CDATA[', '')
            lne = lne.replace(']]', '')
            #lne = lne[1:-1]
        return lne

    @staticmethod
    def getattrs(line, token):
        "split for attributes."
        index = 0
        result = []
        stop = False
        while not stop:
            index1 = line.find(f'<{token} ', index)
            if index1 == -1:
                return result
            index1 += len(token) + 2
            index2 = line.find('/>', index1)
            if index2 == -1:
                return result
            result.append(line[index1:index2])
            index = index2
        return result

    @staticmethod
    def parse(txt, toke="outline", itemz=None):
        "parse on outlines."
        if itemz is None:
            itemz = ",".join(Parser.getnames(txt))
        result = []
        for attrz in Parser.getattrs(txt, toke):
            if not attrz:
                continue
            obj = Default()
            for itm in spl(itemz):
                if itm == "link":
                    itm = "href"
                val = Parser.getvalue(attrz, itm)
                if not val:
                    continue
                if itm == "href":
                    itm = "link"
                setattr(obj, itm, val.strip())
            result.append(obj)
        return result


def attrs(obj, txt):
    "parse attributes into the object."
    update(obj, Parser.parse(txt))


def exp(event):
    "export to opml."
    event.reply(TEMPLATE)
    nrs = 0
    for _fn, objr in broker.all("rss"):
        nrs += 1
        obj = Default()
        update(obj, objr)
        name = obj.name or f"url{nrs}"
        txt = f'<outline name="{name}" display_list="{obj.display_list}" xmlUrl="{obj.rss}"/>'
        event.reply(" "*12 + txt)
    event.reply(" "*8 + "</outline>")
    event.reply("    <body>")
    event.reply("</opml>")


def imp(event):
    "import opml."
    if not event.args:
        event.reply("imp <filename>")
        return
    fnm = event.args[0]
    with open(fnm, "r", encoding="utf-8") as file:
        txt = file.read()
    prs = Parser()
    nrs = 0
    insertid = shortid()
    for obj in prs.parse(txt, 'outline', "name,display_list,xmlUrl"):
        #if obj.xmlUrl and broker.find({"rss": obj.xmlUrl}):
        #    event.reply(f"skipping {obj.xmlUrl}")
        #    continue
        rss = Rss()
        construct(rss, obj)
        rss.rss = rss.xmlUrl
        rss.insertid = insertid
        broker.add(rss)
        nrs += 1
    if nrs:
        event.reply(f"added {nrs} urls.")
        


# This file is placed in the Public Domain.


"slogan"


TXT = """By law, with the use of poison, killed, tortured, castrated, destroyed in whole or in part,
all elderly and all handicapped (Wzd), all criminals (Wfz) and all psychiatric patients (WvGGZ)
here in the Netherlands."""


def lne(event):
    "display line of the slogan."
    event.reply(TXT.split("\n", maxsplit=1)[0])


def slg(event):
    "show slogan."
    event.reply(TXT)



# This file is placed in the Public Domain.
#
# pylint: disable=W0401,W0611,W0614


"interface"


from .classes  import Classes
from .client   import Client, cmnd, scan
from .commands import Commands, command
from .default  import Default
from .disk     import Workdir
from .errors   import Errors, errors, later
from .event    import Event
from .handler  import Handler
from .log      import Logging, debug
from .object   import *
from .parser   import parse
from .repeater import Repeater
from .run      import broker
from .thread   import launch
from .timer    import Timer
from .utils    import *


objxdir = [
        'Object',
        'construct',
        'dump',
        'dumps',
        'edit',
        'fmt',
        'fqn',
        'hook',
        'ident',
        'items',
        'keys',
        'load',
        'loads',
        'read',
        'search',
        'update',
        'values',
        'write'
    ]



objrdir = [
        'Classes',
        'Client',
        'Commands',
        'Errors',
        'Event',
        'Handler',
        'Logging',
        'Repeater',
        'Thread',
        'Timer',
        'Workdir',
        'broker',
        'cmnd',
        'command',
        'debug',
        'errors',
        'init',
        'laps',
        'later',
        'launch',
        'name',
        'scan',
        'spl'
    ]


def __dir__():
    return sorted(objxdir + objrdir)



# This file is placed in the Public Domain.


"disk"


import os
import pathlib
import _thread


from .object  import Object, ident, read, write
from .utils   import strip


lock = _thread.allocate_lock()


class Workdir(Object): # pylint: disable=R0903

    "Workdir"

    workdir = ""

    @staticmethod
    def types():
        "return types stored."
        return os.listdir(Workdir.store())

    @staticmethod
    def skel():
        "create directory,"
        pth  = os.path.join(Workdir.workdir, "store", "")
        path = pathlib.Path(pth)
        path.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def store(pth=""):
        "return objects directory."
        return os.path.join(Workdir.workdir, "store", pth)


def fetch(obj, pth):
    "read object from disk."
    with lock:
        pth2 = Workdir.store(pth)
        read(obj, pth2)
        return strip(pth)


def sync(obj, pth=None):
    "sync object to disk."
    with lock:
        if pth is None:
            pth = ident(obj)
        pth2 = Workdir.store(pth)
        write(obj, pth2)
        return pth


def __dir__():
    return (
        'fetch',
        'sync'
    )



# This file is placed in the Public Domain.


"find"


import os


from .classes import Classes
from .default import Default
from .object  import fqn, search, update
from .disk    import Workdir, fetch
from .utils   import fntime, strip


def fns(mtc=""):
    "show list of files."
    dname = ''
    pth = Workdir.store(mtc)
    for rootdir, dirs, _files in os.walk(pth, topdown=False):
        if dirs:
            for dname in sorted(dirs):
                if dname.count('-') == 2:
                    ddd = os.path.join(rootdir, dname)
                    fls = sorted(os.listdir(ddd))
                    for fll in fls:
                        yield strip(os.path.join(ddd, fll))


def find(mtc, selector=None, index=None, deleted=False):
    "find object matching the selector dict."
    clz = Classes.long(mtc)
    nrs = -1
    result = []
    for fnm in sorted(fns(clz), key=fntime):
        obj = Default()
        fetch(obj, fnm)
        if not deleted and '__deleted__' in dir(obj):
            continue
        if selector and not search(obj, selector):
            continue
        nrs += 1
        if index is not None and nrs != int(index):
            continue
        result.append((fnm, obj))
    return result


def last(obj, selector=None):
    "return last object saved."
    if selector is None:
        selector = {}
    result = sorted(
                    find(fqn(obj), selector),
                    key=lambda x: fntime(x[0])
                   )
    res = None
    if result:
        inp = result[-1]
        update(obj, inp[-1])
        res = inp[0]
    return res


def __dir__():
    return (
        'fns',
        'find',
        'last'
    )



# This file is placed in the Public Domain.


"timer"


import threading
import time


from .object import Object
from .run    import broker
from .thread import launch
from .utils  import named


class Timer(Object):

    "Timer"

    def __init__(self, sleep, func, *args, thrname=None):
        self.args  = args
        self.func  = func
        self.sleep = sleep
        self.name  = thrname or named(func)
        self.state = {}
        self.timer = None
        broker.add(self)

    def run(self):
        "run the payload in a thread."
        self.state["latest"] = time.time()
        launch(self.func, *self.args)

    def start(self):
        "start timer."
        timer = threading.Timer(self.sleep, self.run)
        timer.name   = self.name
        timer.daemon = True
        timer.sleep  = self.sleep
        timer.state  = self.state
        timer.func   = self.func
        timer.state["starttime"] = time.time()
        timer.state["latest"]    = time.time()
        timer.start()
        self.timer   = timer

    def stop(self):
        "stop timer."
        if self.timer:
            self.timer.cancel()


def __dir__():
    return (
        'Timer',
    )



# This file is placed in the Public Domain.


"client"


from .commands import command
from .commands import scan as scancmd
from .errors   import later
from .event    import Event
from .handler  import Handler
from .run      import broker
from .utils    import skip, spl


class Client(Handler):

    "Client"

    def __init__(self):
        Handler.__init__(self)
        self.register("command", command)

    def announce(self, txt):
        "announce text."
        self.raw(txt)

    def raw(self, txt):
        "raw output."

    def say(self, _channel, txt):
        "say text in a channel."
        self.raw(txt)

    def show(self, evt):
        "show results into a channel."
        for txt in evt.result:
            self.say(evt.channel, txt)


def cmnd(txt, outer=None):
    "do a command using the provided output function."
    clt = Client()
    ids = broker.add(clt)
    if outer:
        clt.raw = outer
    evn = Event()
    evn.orig = ids
    evn.txt = txt
    command(clt, evn)
    evn.wait()
    return evn


def init(pkg, modstr, disable=""):
    "start modules"
    mds = []
    for modname in spl(modstr):
        if skip(modname, disable):
            continue
        mod = getattr(pkg, modname, None)
        if not mod:
            continue
        if "init" in dir(mod):
            try:
                mod.init()
                mds.append(mod)
            except Exception as ex: # pylint: disable=W0718
                later(ex)
    return mds


def scan(pkg, modstr, disable=""):
    "scan modules for commands and classes"
    mds = []
    for modname in spl(modstr):
        if skip(modname, disable):
            continue
        module = getattr(pkg, modname, None)
        if not module:
            continue
        scancmd(module)
    return mds


def __dir__():
    return (
        'Client',
        'cmnd',
        'init',
        'scan'
    )



# This file is placed in the Public Domain.


"configuration"


from .default import Default


class Config(Default): # pylint: disable=R0903

    "Config"


def __dir__():
    return (
        'Config',
    )



# This file is placed in the Public Domain.


"client"


import inspect


from .object import Object
from .parser import parse
from .errors import later


class Commands: # pylint: disable=R0903

    "Commands"

    cmds = Object()


    @staticmethod
    def add(func):
        "add command."
        setattr(Commands.cmds, func.__name__, func)


def command(bot, evt):
    "check for and run a command."
    parse(evt)
    func = getattr(Commands.cmds, evt.cmd, None)
    if func:
        try:
            func(evt)
        except Exception as exc: # pylint: disable=W0718
            later(exc)
    bot.show(evt)
    evt.ready()


def scan(mod) -> None:
    "scan module for commands."
    for key, cmd in inspect.getmembers(mod, inspect.isfunction):
        if key.startswith("cb"):
            continue
        if 'event' in cmd.__code__.co_varnames:
            Commands.add(cmd)


def __dir__():
    return (
        'Commands',
        'command',
        'scan'
    )



# This file is placed in the Public Domain.


"rich site syndicate"


import html
import html.parser
import re
import time
import urllib
import urllib.request
import _thread


from urllib.error import HTTPError, URLError
from urllib.parse import quote_plus, urlencode


from ..default  import Default
from ..disk     import sync
from ..find     import last
from ..object   import Object, fmt, update, values
from ..repeater import Repeater
from ..run      import broker
from ..thread   import launch
from ..utils    import fntime, laps, spl


def init():
    "start fetcher."
    fetcher = Fetcher()
    fetcher.start()
    return fetcher


DEBUG = False


fetchlock = _thread.allocate_lock()


class Feed(Default): # pylint: disable=R0903

    "Feed"


class Rss(Default): # pylint: disable=R0903

    "Rss"

    def __init__(self):
        Default.__init__(self)
        self.display_list = 'title,link,author'
        self.rss          = ''


broker.register(Rss)


class Seen(Default): # pylint: disable=R0903

    "Seen"

    def __init__(self):
        Default.__init__(self)
        self.urls = []


class Fetcher(Object):

    "Fetcher"

    def __init__(self):
        self.dosave = False
        self.seen = Seen()
        self.seenfn = None
        broker.add(self)

    @staticmethod
    def display(obj):
        "display object."
        result = ''
        displaylist = []
        try:
            displaylist = obj.display_list or 'title,link'
        except AttributeError:
            displaylist = 'title,link,author'
        for key in displaylist.split(","):
            if not key:
                continue
            data = getattr(obj, key, None)
            if not data:
                continue
            data = data.replace('\n', ' ')
            data = striphtml(data.rstrip())
            data = unescape(data)
            result += data.rstrip()
            result += ' - '
        return result[:-2].rstrip()

    def fetch(self, feed, silent=False):
        "fetch feed."
        with fetchlock:
            counter = 0
            result = []
            for obj in reversed(getfeed(feed.rss, feed.display_list)):
                fed = Feed()
                update(fed, obj)
                update(fed, feed)
                url = urllib.parse.urlparse(fed.link)
                if url.path and not url.path == '/':
                    uurl = f'{url.scheme}://{url.netloc}/{url.path}'
                else:
                    uurl = fed.link
                if uurl in self.seen.urls:
                    continue
                self.seen.urls.append(uurl)
                counter += 1
                if self.dosave:
                    sync(fed)
                result.append(fed)
        if silent:
            return counter
        txt = ''
        feedname = getattr(feed, 'name', None)
        if feedname:
            txt = f'[{feedname}] '
        for obj in result:
            txt2 = txt + self.display(obj)
            for bot in values(broker.objs):
                if "announce" in dir(bot):
                    bot.announce(txt2.rstrip())
        return counter

    def run(self, silent=False):
        "fetch all feeds."
        thrs = []
        for _fn, feed in broker.all('rss'):
            thrs.append(launch(self.fetch, feed, silent, name=f"{feed.rss}"))
        return thrs

    def start(self, repeat=True):
        "start fetcher."
        self.seenfn = last(self.seen)
        if repeat:
            repeater = Repeater(300.0, self.run)
            repeater.start()


class Parser:

    "Parser"

    @staticmethod
    def getitem(line, item):
        "match items."
        lne = ''
        index1 = line.find(f'<{item}>')
        if index1 == -1:
            return lne
        index1 += len(item) + 2
        index2 = line.find(f'</{item}>', index1)
        if index2 == -1:
            return lne
        lne = line[index1:index2]
        lne = cdata(lne)
        return lne.strip()

    @staticmethod
    def getitems(text, token):
        "loop for items."
        index = 0
        result = []
        stop = False
        while not stop:
            index1 = text.find(f'<{token}', index)
            if index1 == -1:
                break
            index1 += len(token) + 2
            index2 = text.find(f'</{token}>', index1)
            if index2 == -1:
                break
            lne = text[index1:index2]
            result.append(lne)
            index = index2
        return result

    @staticmethod
    def parse(txt, toke="item", items='title,link'):
        "parse a text for tokens."
        result = []
        for line in Parser.getitems(txt, toke):
            line = line.strip()
            obj = Default()
            for itm in spl(items):
                val = Parser.getitem(line, itm)
                if val:
                    val = unescape(val.strip())
                    val = val.replace("\n", "")
                    val = striphtml(val)
                    setattr(obj, itm, val)
            result.append(obj)
        return result


def getfeed(url, items):
    "fetch a feed."
    result = [Object(), Object()]
    if DEBUG:
        return result
    try:
        rest = geturl(url)
    except (ValueError, HTTPError, URLError):
        return result
    if rest:
        if url.endswith('atom'):
            result = Parser.parse(str(rest.data, 'utf-8'), 'entry', items) or []
        else:
            result = Parser.parse(str(rest.data, 'utf-8'), 'item', items) or []
    return result


def gettinyurl(url):
    "fetch a tinyurl."
    postarray = [
        ('submit', 'submit'),
        ('url', url),
    ]
    postdata = urlencode(postarray, quote_via=quote_plus)
    req = urllib.request.Request('http://tinyurl.com/create.php',
                  data=bytes(postdata, 'UTF-8'))
    req.add_header('User-agent', useragent("rss fetcher"))
    with urllib.request.urlopen(req) as htm: # nosec
        for txt in htm.readlines():
            line = txt.decode('UTF-8').strip()
            i = re.search('data-clipboard-text="(.*?)"', line, re.M)
            if i:
                return i.groups()
    return []


def cdata(line):
    "retrieve text from CDATA."
    if 'CDATA' in line:
        lne = line.replace('![CDATA[', '')
        lne = lne.replace(']]', '')
        lne = lne[1:-1]
        return lne
    return line


def geturl(url):
    "fetch a url."
    if not url.startswith("http://") and not url.startswith("https://"):
        return ""
    url = urllib.parse.urlunparse(urllib.parse.urlparse(url))
    req = urllib.request.Request(url)
    req.add_header('User-agent', useragent("rss fetcher"))
    with urllib.request.urlopen(req) as response: # nosec
        response.data = response.read()
        return response


def striphtml(text):
    "strip html."
    clean = re.compile('<.*?>')
    return re.sub(clean, '', text)


def unescape(text):
    "unescape text."
    txt = re.sub(r'\s+', ' ', text)
    return html.unescape(txt)


def useragent(txt):
    "return useragent."
    return 'Mozilla/5.0 (X11; Linux x86_64) ' + txt


def dpl(event):
    "set display items."
    if len(event.args) < 2:
        event.reply('dpl <stringinurl> <item1,item2>')
        return
    setter = {'display_list': event.args[1]}
    for _fn, feed in broker.find("rss", {'rss': event.args[0]}):
        if feed:
            update(feed, setter)
    event.reply('ok')


def nme(event):
    "set name of feed."
    if len(event.args) != 2:
        event.reply('nme <stringinurl> <name>')
        return
    selector = {'rss': event.args[0]}
    for _fn, feed in broker.find("rss", selector):
        if feed:
            feed.name = event.args[1]
    event.reply('ok')


def rem(event):
    "remove a feed."
    if len(event.args) != 1:
        event.reply('rem <stringinurl>')
        return
    for _fnm, feed in broker.all("rss"):
        if event.args[0] not in feed.rss:
            continue
        if feed:
            feed.__deleted__ = True
    event.reply('ok')


def res(event):
    "restore a feed."
    if len(event.args) != 1:
        event.reply('res <stringinurl>')
        return
    for fnm, feed in broker.all("rss", True):
        if event.args[0] not in feed.rss:
            continue
        if feed:
            feed.__deleted__ = False
            sync(feed, fnm)
    event.reply('ok')


def rss(event):
    "add a feed."
    if not event.rest:
        nrs = 0
        for fnm, feed in broker.all('rss'):
            nrs += 1
            elp = laps(time.time()-fntime(fnm))
            txt = fmt(feed)
            event.reply(f'{nrs} {txt} {elp}')
        if not nrs:
            event.reply('no rss feed found.')
        return
    url = event.args[0]
    if 'http' not in url:
        event.reply('i need an url')
        return
    for fnm, result in broker.find("rss", {'rss': url}):
        if result:
            event.reply(f'already got {url}')
            return
    feed = Rss()
    feed.rss = event.args[0]
    sync(feed)
    event.reply('ok')


def syn(event):
    "synchronize feeds."
    fetchers = list(broker.all("fetcher"))
    if not fetchers:
        event.reply("no fetcher found.")
        return
    fetcher = fetchers[0][-1]
    thrs = fetcher.run(True)
    nrs = 0
    for thr in thrs:
        thr.join()
        nrs += 1
    event.reply(f"{nrs} feeds synced")
 